// new lib based on http://www.mikrocontroller.net/articles/AVR_FAT32 0.6.4 by Daniel R.
#include <ctype.h>
#include <stdint.h>
#include <avr/io.h>
#include <avr/interrupt.h>

#include "types.h"
#include "mrmidi2.h"
#include "delay.h"
#include "spi.h"
#include "mmc.h"

	//#####################################################################################
	// BOOL DEFINITIONEN

	#define TRUE 	0x01
	#define FALSE 	0x00

	//#####################################################################################
	// FLAGS

	// flags fuer ffopen (sind die rueckgabewerte von ffopen, auf diese wird geprueft..)
	#define MMC_FILE_OPENED 	0	// datei existierte und wurde geoeffnet.
	#define MMC_FILE_CREATED 	1	// datei angelegt.
	#define MMC_FILE_ERROR		2	// datei existiert nicht/nicht angelegt.


	//#####################################################################################
	// WICHTIGE SCHALTER

	// schalter die den funktionsumfang aendern bzw, den funktionsumfang einiger funktionen :)
	#define MMC_WRITE 			TRUE	// TRUE, dann mit write unterstuetzung, wenn FALSE dann read only !
	#define MMC_OVER_WRITE 		TRUE	// TRUE und MMC_WRITE TRUE, dann kann ffwrite dateien ueberschreiben, wenn FALSE dann nur normales schreiben. um an eine datei anzuhaengen ist MMC_OVER_WRITE TRUE nicht noetig!
	#define MMC_MULTI_BLOCK 	FALSE	// TRUE und MMC_OVER_WRITE FALSE, dann werden multiblock schreib/lese funktionen benutzt. ist schneller, wird aber moeglicherweise nicht von allen karten untersttzt. wenn FALSE ist normale operation
	#define MMC_ENDIANNESS_LITTLE 	TRUE	// TRUE, dann ist der code auf littleendian ausgelegt. AVR ist littleendian. code ist auf littleendian optimiert!! siehe: http://de.wikipedia.org/wiki/Endianness
	#define MMC_LFN_SUPPORT		TRUE	// TRUE, dann mit unterstuetzung fuer lange dateinamen. kostet wenn read und write benutzt wird um die 800 bytes flash...
	#define MMC_RM_FILES_ONLY	TRUE	// TRUE ,MMC_WRITE TRUE und MMC_RM TRUE, dann wird die funktion ffrm so mit kompiliert, dass sie nur dateien loeschen kann. wenn FALSE und MMC_WRITE TRUE, dann kann die funktion dateien und ordner rekursiv loeschen !

	// schalter die explizit funktionen mit kompilieren oder nicht!
	#define MMC_TIME_STAMP 		FALSE 	// TRUE, dann werden die funktionen fat_getTime und fat_getFreeBytes mit kompiliert. siehe auch abschnitt: ZEIT FUNKTIONEN, weiter unten
	#define MMC_RM 				FALSE	// TRUE und MMC_WRITE TRUE, dann wird die funktion ffrm mit kompiliert.
	#define MMC_SEEK			TRUE	// TRUE,dann wird die funktion ffseek mit kompiliert. mit dieser funktion kann man in einer geoeffneten datei vor und zurueck spulen. nur in kombination mit MMC_OVER_WRITE TRUE kann in einer datei ueberschrieben werden.
	#define MMC_MKDIR			FALSE	// TRUE und MMC_WRITE TRUE, dann wird die funktion ffmkdir mit kompiliert. mit dieser funktion kann man ordner anlegen.
	#define MMC_GET_FREE_BYTES	FALSE	// TRUE, dann wird die funkton fat_getFreeBytes mit kompiliert. mit dieser funktion kann der freie platzt auf der karte ermittelt werden
	#define MMC_LS				TRUE	// TRUE, dann wird die funktion ffls mit kompiliert. mit dieser funkion kann man die dateien auf der karte anzeigen lassen
	#define MMC_CD				TRUE	// TRUE, dann werden die funktionen ffcd und ffcdLower mit kompiliert. mit diesen funktionen kann man in ein verzeichnis wechseln oder aus einem verzeichnis ein verzeichnis hoeher wechseln.
	#define MMC_FILE_EXSISTS	TRUE	// TRUE, dann wird die funktion ffileExsists mit kompiliert. mit dieser funktion kann geprueft werden, ob es die datei im aktuellen verzeinis gibt !
	#define MMC_WRITE_STRING	TRUE	// TRUE und MMC_WRITE TRUE, dann wird die funktion ffwrites mit kompiliert. mit dieser funktion koennen strings auf die karte geschrieben werden.
	#define MMC_WRITEN			FALSE	// TRUE	und MMC_WRITE TRUE, dann wird die funktion ffwriten mit kompiliert.	mit dieser funktion koennen mehrere zeichen auf die karte geschrieben werden!

	// vorsicht, da die variable die die sektoren zaehlt ein short ist (MMC_MAX_CLUSTERS_IN_ROW*fat.secPerClust) !!
	#define MMC_MAX_CLUSTERS_IN_ROW 256 	// gibt an wie viele cluster am stueck ohne fat-lookup geschrieben bzw gelesen werden knnen, wenn die fat nicht fragmentiert ist !


	//#####################################################################################
	// SPI EINSTELLUNGEN

	// wenn MAX_SPEED FALSE dann wird die SPI bus geschwindigkeit nicht hoch gesetzt. ist zum testen von hardware die nicht richtig luft
	#define MMC_MAX_SPEED 		TRUE
	#define MMC_STATUS_INFO 	FALSE

	// software spi? wenn TRUE muessen in der mmc.h noch die pins eingestellt werden ueber die die spi kommunikation laufen soll
	// es werden nur pins an einem port unterstuetzt !
	#define MMC_SOFT_SPI 	FALSE


	#define MMC_Write 		PORTB	//Port an der die MMC/SD-Karte angeschlossen ist also des SPI
	#define MMC_Read 		PINB
	#define MMC_Direction_REG DDRB

	#if defined (__AVR_ATmega328__) || defined (__AVR_ATmega328P__) || defined (__AVR_ATmega168__) || defined (__AVR_ATmega168P__)
		#define SPI_MISO 			4  //Port Pin an dem Data Output der MMC/SD-Karte angeschlossen ist
		#define SPI_MOSI    		3  //Port Pin an dem Data Input der MMC/SD-Karte angeschlossen ist
		#define SPI_Clock 			5  //Port Pin an dem die Clock der MMC/SD-Karte angeschlossen ist (clk)
		#define SPI_SS    			2  //Port Pin an dem Chip Select der MMC/SD-Karte angeschlossen ist (CS)
	#else
		#error "unknown MCU"
	#endif




// Definitions for MMC/SDC command
#define CMD0	(0)			// GO_IDLE_STATE
#define CMD1	(1)			// SEND_OP_COND (MMC)
#define	ACMD41	(41)		// SEND_OP_COND (SDC)
#define CMD8	(8)			// SEND_IF_COND
#define CMD9	(9)			// SEND_CSD
#define CMD10	(10)		// SEND_CID
#define CMD12	(12)		// STOP_TRANSMISSION
#define ACMD13	(0x80+13)	// SD_STATUS (SDC)
#define CMD16	(16)		// SET_BLOCKLEN
#define CMD17	(17)		// READ_SINGLE_BLOCK
#define CMD18	(18)		// READ_MULTIPLE_BLOCK
#define CMD23	(23)		// SET_BLOCK_COUNT (MMC)
#define	ACMD23	(0x80+23)	// SET_WR_BLK_ERASE_COUNT (SDC)
#define CMD24	(24)		// WRITE_BLOCK
#define CMD25	(25)		// WRITE_MULTIPLE_BLOCK
#define CMD55	(55)		// APP_CMD
#define CMD58	(58)		// READ_OCR

// Card type flags (CardType)
#define CT_MMC		0x01			// MMC ver 3
#define CT_SD1		0x02			// SD ver 1
#define CT_SD2		0x04			// SD ver 2
#define CT_SDC		(CT_SD1|CT_SD2)	// SD
#define CT_BLOCK	0x08			// Block addressing

// **********************************************************************************************************************************
// funktionsprototypen von funktionen die nur in dieser datei benutzt werden !

static uint8_t 	mmc_enable(void);
static void 	mmc_disable(void);
static uint8_t 	mmc_wait_ready (void);
static uint8_t 	mmc_send_cmd (	uint8_t cmd,	uint32_t arg);

// beginn -> hardware abhaengiger teil !
#define MMC_CS_LOW 		MMC_Write &= ~(1<<SPI_SS)		// Set pin B2 to 0
#define MMC_CS_HIGH		MMC_Write |= (1<<SPI_SS)		// Set pin B2 to 1

static void 	spi_init_mmc(void);
static void 	spi_maxSpeed(void);
static void 	spi_write_byte(uint8_t byte);
static uint8_t 	spi_read_byte(void);

uint8_t 		card_type;

//uint8_t mmc_get_card_type(){return card_type;}

// *****************************************************************************
static void spi_init_mmc(void){

	// port configuration der mmc/sd/sdhc karte
	MMC_Direction_REG &=~(1<<SPI_MISO);         // miso auf input
	MMC_Direction_REG |= (1<<SPI_Clock);      	// clock auf output
	MMC_Direction_REG |= (1<<SPI_MOSI);         // mosi auf output
	MMC_Direction_REG |= (1<<SPI_SS);			// chip select auf output

	// hardware spi: bus clock = idle low, spi clock / 128 , spi master mode
	SPCR = (1<<SPE)|(1<<MSTR)|(1<<SPR0)|(1<<SPR1);

	MMC_Write |= (1<<SPI_SS);       	// chip selet auf high, karte anwaehlen
}



#if (MMC_MAX_SPEED==TRUE)
// *****************************************************************************
static void spi_maxSpeed(){

	//SPI Bus auf max Geschwindigkeit
	SPCR &= ~((1<<SPR0) | (1<<SPR1));
	SPSR |= (1<<SPI2X);
}
#endif

// *****************************************************************************
static void spi_write_byte(uint8_t byte){


#if (MMC_SOFT_SPI==TRUE)
	uint8_t a;
#endif

	// mmc/sd in hardware spi
#if (MMC_SOFT_SPI==FALSE)
	SPDR = byte;    						//Sendet ein Byte
	loop_until_bit_is_set(SPSR,SPIF);

	// mmc/sd in software spi
#else
	for (a=8; a>0; a--){					//das Byte wird Bitweise nacheinander Gesendet MSB First
		if (bit_is_set(byte,(a-1))>0){		//Ist Bit a in Byte gesetzt
			MMC_Write |= (1<<SPI_MOSI); 	//Set Output High
		}
		else{
			MMC_Write &= ~(1<<SPI_MOSI); 	//Set Output Low
		}
		MMC_Write |= (1<<SPI_Clock); 		//setzt Clock Impuls wieder auf (High)
		MMC_Write &= ~(1<<SPI_Clock);		//erzeugt ein Clock Impuls (LOW)
	}
	MMC_Write |= (1<<SPI_MOSI);				//setzt Output wieder auf High
#endif
}


// *****************************************************************************
static uint8_t spi_read_byte(void){

	// mmc/sd in hardware spi
#if (MMC_SOFT_SPI==FALSE)
	SPDR = 0xff;
	loop_until_bit_is_set(SPSR,SPIF);
	return (SPDR);

	// mmc/sd in software spi
#else
	uint8_t Byte=0;
	uint8_t a;
	for (a=8; a>0; a--){							//das Byte wird Bitweise nacheinander Empangen MSB First
		MMC_Write |=(1<<SPI_Clock);					//setzt Clock Impuls wieder auf (High)
		if (bit_is_set(MMC_Read,SPI_MISO) > 0){ 	//Lesen des Pegels von MMC_MISO
			Byte |= (1<<(a-1));
		}
		else{
			Byte &=~(1<<(a-1));
		}
		MMC_Write &=~(1<<SPI_Clock); 				//erzeugt ein Clock Impuls (Low)
	}
	return (Byte);
#endif
}



// ende <- hardware abhaengiger teil !








// **********************************************************************************************************************************
uint8_t mmc_init (void){

	uint8_t cmd, ty, ocr[4];
	uint16_t n, j;

	for (n = 100; n; n--) spi_read_byte();    					// 80+ dummy clocks

	ty = 0;
	j=100;
	do {
		if (mmc_send_cmd(CMD0, 0) == 1) {      					// Enter Idle state
			j=0;
			u16up TimingDelayLocal = 1000;  					// Initialization timeout

			if (mmc_send_cmd(CMD8, 0x1AA) == 1) {  				// SDv2?
				for (n = 0; n < 4; n++){
					ocr[n] = spi_read_byte();    				// Get trailing return value of R7 resp
				}
				if (ocr[2] == 0x01 && ocr[3] == 0xAA) {         // The card can work at vdd range of 2.7-3.6V
					while (TimingDelayLocal--) {  						// Wait for leaving idle state (ACMD41 with HCS bit)
						mmc_send_cmd(CMD55, 0);
						if(!mmc_send_cmd(ACMD41, 1UL << 30))
						break;
					}

					while(TimingDelayLocal--) {
						if (mmc_send_cmd(CMD58, 0) == 0x00) {    // Check CCS bit in the OCR
							for (n = 0; n < 4; n++){
								ocr[n] = spi_read_byte();
							}
							ty = (ocr[0] & 0x40) ? CT_SD2 | CT_BLOCK : CT_SD2;  // SDv2
							break;
						}
					}
				}
				} else {        									// SDv1 or MMCv3
				if (mmc_send_cmd(ACMD41, 0) <= 1)   {
					ty = CT_SD1;
					cmd = ACMD41;  								// SDv1
					} else {
					ty = CT_MMC;
					cmd = CMD1;    								// MMCv3
				}
				while (TimingDelayLocal-- && mmc_send_cmd(cmd, 0));    // Wait for leaving idle state
			}
			if(ty != (CT_SD2 | CT_BLOCK)) {
				while(TimingDelayLocal-- && (mmc_send_cmd(CMD16, 512) != 0));
			}
			if(!TimingDelayLocal) ty = 0;
			} else { j--; }
		}while(j>0);

		card_type = ty;
		mmc_disable();

		if( card_type == 0 ){
			return FALSE;
		}
	#if (MMC_MAX_SPEED==TRUE)
		spi_maxSpeed();
	#endif

		return TRUE;
	}

	// **********************************************************************************************************************************
	static uint8_t mmc_send_cmd (	uint8_t cmd,	uint32_t arg){

		uint8_t n, res;
		// Select the card and wait for ready
		mmc_disable();
		if ( FALSE == mmc_enable() ){
			return 0xFF;
		}
		// Send command packet
		spi_write_byte(0x40 | cmd);						// Start + Command index
		spi_write_byte( (uint8_t)(arg >> 24) );	// Argument[31..24]
		spi_write_byte( (uint8_t)(arg >> 16) );	// Argument[23..16]
		spi_write_byte( (uint8_t)(arg >> 8) );	// Argument[15..8]
		spi_write_byte( (uint8_t)arg );			// Argument[7..0]
		n = 0x01;										// Dummy CRC + Stop
		if (cmd == CMD0) n = 0x95;						// Valid CRC for CMD0(0)
		if (cmd == CMD8) n = 0x87;						// Valid CRC for CMD8(0x1AA)
		spi_write_byte(n);

		// Receive command response
		if (cmd == CMD12) spi_read_byte();				// Skip a stuff byte when stop reading
		n = 10;											// Wait for a valid response in timeout of 10 attempts
		do
		res = spi_read_byte();
		while ( (res & 0x80) && --n );

		return res;										// Return with the response value
	}





	// **********************************************************************************************************************************
	static uint8_t mmc_enable(){

		MMC_CS_LOW;
		if( !mmc_wait_ready() ){
			mmc_disable();
			return FALSE;
		}

		return TRUE;
	}

	// **********************************************************************************************************************************
	static void mmc_disable(){

		MMC_CS_HIGH;
		spi_read_byte();
	}



	// **********************************************************************************************************************************
	// wartet darauf, dass die mmc karte in idle geht
	// **********************************************************************************************************************************
	static uint8_t mmc_wait_ready (void){

		u16up TimingDelayLocal = 5000;

		do{
			if(	 spi_read_byte() == 0xFF ) return TRUE;
		}while ( TimingDelayLocal-- );

		return FALSE;
	}




	// **********************************************************************************************************************************
	// Routine zum schreiben eines Blocks(512Byte) auf die MMC/SD-Karte
	// **********************************************************************************************************************************
	uint8_t mmc_write_sector (uint32_t addr,uint8_t *buffer){

		uint8_t resp;
		uint8_t retrys;
		uint16_t count;

		if ( !(card_type & CT_BLOCK) ){
			addr *= 512;				// Convert to byte address if needed
		}

		if ( mmc_send_cmd(CMD24, addr) != 0){ 	// enables card
			return FALSE;
		}

		if ( FALSE == mmc_wait_ready() ){
			return FALSE;
		}

		spi_write_byte(0xFE);			// Xmit data token

		count = 512;
		do {							// Xmit the 512 byte data block to MMC
			spi_write_byte(*buffer++);
		} while (--count);

		spi_write_byte(0xFF);			// CRC (Dummy)
		spi_write_byte(0xFF);

		retrys = 20;
		do{
			resp = spi_read_byte();		// Reveive data response, 20 retrys if not acepted
		}while( (resp & 0x1F) != 0x05 && --retrys);

		if ( retrys == 0){				// If not accepted, return with error
			return FALSE;
		}

		mmc_disable();

		return TRUE;
	}


	// **********************************************************************************************************************************
	// Routine zum lesen eines Blocks(512Byte) von der MMC/SD-Karte
	// **********************************************************************************************************************************
	uint8_t mmc_read_sector (uint32_t addr,uint8_t *buffer){

		uint8_t token;
		uint16_t count;
		u16up TimingDelayLocal = 2000;

		if ( !(card_type & CT_BLOCK) ) addr *= 512;	// Convert to byte address if needed

		if ( mmc_send_cmd(CMD17, addr) != 0 ){
			return FALSE;
		}

		do {							// Wait for data packet in timeout of 200ms
			token = spi_read_byte();
		} while ( (token == 0xFF) && TimingDelayLocal-- );

		if(token != 0xFE){
			return FALSE;				// If not valid data token, retutn with error
		}

		count = 512;
		do {							// Receive the data block into buffer
			*buffer++ = spi_read_byte();
		} while (--count);

		spi_read_byte();				// Discard CRC
		spi_read_byte();

		mmc_disable();

		return TRUE;					// Return with success
	}

// end new lib




// simon stuff

void MMC_hw_init(void)
{
	spi_init_mmc();
	mmc_disable();
}

// starts up the MMC card
u08up MMC_init(void)
{
	return mmc_init();
}


static void MMC_write_sec_start(u32up addr)
{
	if ( !(card_type & CT_BLOCK) ){
		addr *= 512;				// Convert to byte address if needed
	}

	if ( mmc_send_cmd(CMD24, addr) != 0){ 	// enables card
		return FALSE;
	}

	if ( FALSE == mmc_wait_ready() ){
		return FALSE;
	}

	spi_write_byte(0xFE);			// Xmit data token
}

static void MMC_get_sec_stop_w(void)
{
	uint8_t resp;
	uint8_t retrys;

	spi_write_byte(0xFF);			// CRC (Dummy)
	spi_write_byte(0xFF);

	retrys = 20;
	do{
		resp = spi_read_byte();		// Receive data response, 20 retrys if not accepted
	}while( (resp & 0x1F) != 0x05 && --retrys);

	if ( retrys == 0){				// If not accepted, return with error
		return FALSE;
	}

	mmc_disable();
}


static void MMC_get_sec_start(u32up addr)
{
	uint8_t token;
	u32up TimingDelayLocal = 20000;

	if ( !(card_type & CT_BLOCK) ) addr *= 512;	// Convert to byte address if needed

	if ( mmc_send_cmd(CMD17, addr) != 0 ){
		return FALSE;
	}

	do {							// Wait for data packet
		token = spi_read_byte();
	} while ( (token == 0xFF) && --TimingDelayLocal );
}

static void MMC_get_sec_stop(void)
{
	spi_read_byte();				// Discard CRC
	spi_read_byte();

	mmc_disable();
}



//-------------------------------------------------------------
// High Level MMC-Funktionen

#include "fat16.h"

u32 sect;
u16 bufpos;

extern u08 flags, fat_directaccess;

// Lesen beginnen
void mmc_load_start(u32up s) {
	sect = s;
	MMC_get_sec_start(s);
	bufpos = 0;
}

// Byte lesen
u08up mmc_fetch_byte(void) {
	if (bufpos == 512) {
		MMC_get_sec_stop();
		if (!fat_directaccess)
			fat_getnextsector_rd();
		else
			sect++;
		mmc_load_start(sect);
	}
	bufpos++;
	return spi_read_byte();
}

// Lesen, um die 512 voll zu machen
void mmc_complete_read(void) {
	while (bufpos < 512) {
		spi_read_byte();
		bufpos++;
	}
	MMC_get_sec_stop();
}

void mmc_read_block(u08 *buf, u16up cnt) {
	while (cnt--) {
		if (bufpos == 512) {
			MMC_get_sec_stop();
			if (!fat_directaccess)
				fat_getnextsector_rd();
			else
				sect++;
			mmc_load_start(sect);
		}
		bufpos++;
		*buf++ = spi_read_byte();
	}
}


#ifndef NO_RECORD

// Schreiben beginnen
void mmc_write_start(u32up s) {
	sect = s;
	MMC_write_sec_start(s);
	bufpos = 0;
}

// Byte schreiben
void mmc_write_byte(u08up b) {
	if (bufpos == 512) {
		MMC_get_sec_stop_w();
		if (!fat_directaccess)
			fat_getnextsector_wr();
		else
			sect++;
		mmc_write_start(sect);
	}
	bufpos++;
	spi_write_byte(b);
}

// Schreiben, um die 512 voll zu machen
void mmc_complete_write(void) {
	while (bufpos < 512) {
		spi_read_byte();
		bufpos++;
	}
	MMC_get_sec_stop_w();
}

void mmc_write_block(u08 *buf, u16up cnt) {
	while (cnt--) {
		mmc_write_byte(*buf);
		buf++;
	}
}

#endif

